/* HTTP server
 * responds to GET / HTTP/1.0 with root.htm
 * responds to GET /0c:/filename.ext|/ HTTP/1.0 with the file
 * no binary transfers or searches
 * based entirely on Chris McNeil's Gopher server
 * ashok@biochemistry.cwru.edu
 *
 * Maintain Gopher item types.  Allows to add search and binary
 * file retrieval later on.
 */

/* Above comments of Ashok don't apply anymore.
   responds to GET /path/filename.ext HTTP/1.0 or,
	       GET /drive:/path/filename.ext HTTP/1.0
   If it's a directory, responds with welcome.htm or
   a directory menu is prepared on the fly.
   No searches, but, everything is transfered in binary mode.

   Selcuk Ozturk // seost2+@pitt.edu // March 31, 1995
 */

/* UMN's wonderful Gopher Server */
/* Coded by Chris McNeil, Mount Allison University */
/* cmcneil@mta.ca  */
/* Parts of code taken from Fingerd.c  */


/* Copyright 1991 Phil Karn, KA9Q */

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <dir.h>
#include <dos.h>
#include <ctype.h>
#include "global.h"
#include "files.h"
#include "mbuf.h"
#include "socket.h"
#include "session.h"
#include "proc.h"
#include "dirutil.h"
#include "commands.h"
#include "mailbox.h"
#include "netuser.h"
#include "ftp.h"
#include "smtp.h"
#include "alloc.h"
#include "cmdparse.h"


int HttpUsers;
static char entry[] = "%-6s<a href=\"/%c:%s%s\">%13s</a>%15ld bytes   %2d %s 19%d\n";

extern char *Months[];


/* function prototypes */
static void http __ARGS((int s,void *unused,void *p));
static int mkwelcome __ARGS((int s,char *Hdir,char *file));
static int dohttptdisc __ARGS((int argc,char *argv[],void *p));
static int dohttpsim __ARGS((int argc,char *argv[],void *p));
static int dohttpmax __ARGS((int argc,char *argv[],void *p));
static int dohttpdrive __ARGS((int argc,char *argv[],void *p));

extern int findport(int s, char *dir, char c);

static struct cmds Httpcmds[] = {
  "drive",       dohttpdrive,   0, 0, NULLCHAR,
  "maxcli",      dohttpmax,     0, 0, NULLCHAR,
  "simult",      dohttpsim,     0, 0, NULLCHAR,
  "tdisc",       dohttptdisc,   0, 0, NULLCHAR,
  NULLCHAR,
};

int
dohttp(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    return subcmd(Httpcmds,argc,argv,p);
}

static char DR = 'c';

static int
dohttpdrive(argc,argv,p)
int argc;
char *argv[];
void *p;
{
  if(argc < 2)
    tprintf("Def. drive: %c\n",DR);
  else if(!isalpha(*argv[1]))
    tputs("Bad drive letter\n");
  else
    DR = *argv[1];
  return 0;
}

int Httpmax = 10;

static int
dohttpmax(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    return setint(&Httpmax,"Max. connects",argc,argv);
}

int Httpsimc = 5;

static int
dohttpsim(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    return setint(&Httpsimc,"Simult. conn.'s serviced",argc,argv);
}

int Httptdisc = 180;

static int
dohttptdisc(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    return setint(&Httptdisc,"Server tdisc (sec)",argc,argv);
}

/* Start up http service */
int httpstart(int argc, char *argv[], void *p)
{

	int16 port;

	if(argc < 2)
	  port = IPPORT_HTTP;
	else
	  port = atoi(argv[1]);

	return start_tcp(port,"HTTP Server",http,1024);

      }


/* Creates a copy of string which has '%' escaped charecters decoded
   back to ASCII. Allocates memory. Caller must free it. */

char *decode(char *string)
{
  int i,k,slen;
  char code[3],*tmp;

  slen = strlen(string);
  tmp = mallocw(slen + 1);
  i = k = 0;

  while(i <= slen) {
    if(string[i] == '%') {
      string[i] = '#';        /* '%' sign causes problems in log()  */
      code[0] = string[i+1];
      code[1] = string[i+2];
      code[2] = '\0';
      tmp[k] = (char)htoi(code);
      k++;  i += 3;
    } else
      tmp[k++] = string[i++];
  }
  return tmp;
}

static void http(int s, void *unused, void *p)
{
	struct sockaddr fsocket;
	char command[PLINELEN], *cp, *newcommand, *tmp;
	FILE *fp;
	char file[PLINELEN+16],Hdir[SLINELEN];
	int port, length, i, err = 1;
	


	sockowner(s, Curproc);
	sockmode(s, SOCK_ASCII);
	HttpUsers++;

	if(availmem() < Memthresh)
	  goto quit;
	if(Httpmax < HttpUsers)
	  goto quit;               

	while(HttpUsers > Httpsimc)
	  pwait(&HttpUsers);

	alarm(Httptdisc * 1000L);
	if(recvline(s, command, PLINELEN) == -1)
	  goto quit;
	alarm(0L);

	port = findport(s,Hdir,'w');

	rip(command);
	tmp = decode(command);
	cp = strtok(tmp,"/");
	newcommand = strtok(NULL," ");
	
	/* newcommand will be everything between "/" and " " */
	length = strlen(newcommand);

#ifdef notdef
	if(strnicmp(command,"POST",4) == 0){
	  log(s,command);
	  while(recvline(s,file,sizeof(file)) != -1) {
	    rip(file);
	    log(s,file);
	  }
	  free(tmp);
	  goto quit;
	}
#endif
	
	if(!strnicmp(newcommand,"b:",2))        
		newcommand[0] = DR;

	if ((strstr(newcommand, "HTTP") != NULL) || 
	    (length == 0) || !strcmp(&newcommand[1],":/"))
						/* for "GET / HTTP/V1.0" */
	{
		sprintf(file,"%c:%s/root.html",
			(newcommand[1] == ':') ? newcommand[0] : DR,Hdir);

	} 
	else if(newcommand[1] == ':') {
	  sprintf(file,"%c:%s%s",newcommand[0],Hdir,&newcommand[2]);      
	} else {
	  sprintf(file,"%c:%s/%s",DR,Hdir,newcommand);
	}

	/* delete any .. to avoid */
	/* anyone telnetting to http and cd'ing to a directory */
	/* other than Httpdir */

	for ( i=0; i <= strlen(file); i++ ) {
	  if (file[i] == '.' && file[i+1] == '.') {
	    file[i] = ' ';
	    file[i+1] = ' ';
	  }
	}

	if(file[strlen(file)-1] == '/') {
	  strcat(file,"welcome.htm");
	  if((fp = fopen(file,READ_BINARY)) == NULLFILE)
	    err = mkwelcome(s,Hdir,file);
	} else if((fp = fopen(file, READ_BINARY)) == NULLFILE &&
		  !access(file,0)) {
	  usprintf(s,"HTTP/1.0 302 Found\nLocation: http://%s/%s/\n\n",
		   Hostname,newcommand);
	  usprintf(s,"<TITLE>Redirection</TITLE><H1>Redirection</H1>Please,"
		   "\n<A HREF=\"http://%s/%s/\">click here</A>\n",
		   Hostname,newcommand);
	}

	if (fp != NULLFILE){
	  sendfile(fp, s, IMAGE_TYPE, 0, NULL);
	  fclose(fp);
	} else if(err)
	  usprintf(s, "<TITLE>Error</TITLE><H1>404 Not Found</H1>The file \"/%s\" does not exist on this server.\n",newcommand);
	
	free(tmp);
	log(s,"HTTP #%u: %s",port,command);
 
quit:
	close_s(s);
	HttpUsers--;
	psignal(&HttpUsers,0);
}


int http0(int argc, char *argv[], void *p)
{
  int16 port;

  if(argc < 2)
    port = IPPORT_HTTP;
  else
    port = atoi(argv[1]);

  return stop_tcp(port);
}

static 
int mkwelcome(int s,char *Hdir,char *file)
{
  char *dirstring,*p,*cp;
  FILE *fp;                
  struct ffblk ffblk;
  int done,pre = 0;

  if((dirstring = (char *)mallocw(strlen(file) + 16)) == NULL)
    return 0;

  strcpy(dirstring,file);
  cp = strrchr(dirstring,'/'); 
  *++cp = '\0';

  usputc(s,'\n');   /* Netscape seems to need this, why? */
  strcat(dirstring,"wwwpre.cat");
  if((fp = fopen(dirstring,READ_BINARY)) != NULLFILE) {
    sendfile(fp,s,0,IMAGE_TYPE,NULL);
    fclose(fp);
    pre = 1;
  }

  *cp = '\0';
  strcat(dirstring,"*.*");
  done = findfirst(dirstring,&ffblk,FA_DIREC);

  sockmode(s,SOCK_ASCII);
  setflush(s,-1); /* we will do our own flushing */
  if(!done) {
    *cp = '\0';
    p = dirstring + strlen(Hdir) + 2;
    undosify(p);
    if(!pre)
      usprintf(s,"<TITLE>Dir</TITLE><H2>Directory of \"%c:%s\"</H2>\n",file[0],p);
    usputs(s,"<PRE>\n");
  } else {
    free(dirstring);
    return -1;
  }

  while(!done) {
    strlwr(ffblk.ff_name);
    if(ffblk.ff_name[0] != '.') {
      if(ffblk.ff_attrib == FA_DIREC)
	cp = "[DIR]";
      else
	cp = "[FILE]";
      usprintf(s,entry,cp,file[0],p,ffblk.ff_name,ffblk.ff_name,ffblk.ff_fsize,
	       (ffblk.ff_fdate) & 0x1f, /* day */
	       Months[((ffblk.ff_fdate >> 5) & 0xf)-1], /* month */
	       (ffblk.ff_fdate >> 9) + 80); /* year */
    }
    pwait(NULL);
    done = findnext(&ffblk);
  }
  usputs(s,"</PRE>\n");
  usflush(s);

  strcat(dirstring,"wwwpost.cat");
  fp = fopen(dirstring,READ_BINARY);
  if(fp != NULLFILE) {
    sendfile(fp,s,0,IMAGE_TYPE,NULL);
    fclose(fp);
  }
  free(dirstring);
  return 0;
}
